@@ -16,7 +16,7 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#ident "$Id: iso2022.c,v 1.43 2003/05/21 21:42:16 nalin Exp $"
+#ident "$Id: iso2022.c,v 1.52 2005/02/28 21:32:01 kmaraas Exp $"
 #include "../config.h"
 #include <sys/types.h>
 #include <errno.h>
@@ -298,28 +298,35 @@ _vte_iso2022_is_ambiguous(gunichar c)
 {
 	int i;
 	gpointer p;
-	static GTree *ambiguous = NULL;
+	static GHashTable *ambiguous = NULL;
 	for (i = 0; i < G_N_ELEMENTS(_vte_iso2022_ambiguous_ranges); i++) {
 		if ((c >= _vte_iso2022_ambiguous_ranges[i].start) &&
 		    (c <= _vte_iso2022_ambiguous_ranges[i].end)) {
 			return TRUE;
 		}
 	}
-	if (ambiguous == NULL) {
-		ambiguous = g_tree_new(_vte_direct_compare);
-		for (i = 0;
-		     i < G_N_ELEMENTS(_vte_iso2022_ambiguous_chars);
-		     i++) {
-			p = GINT_TO_POINTER(_vte_iso2022_ambiguous_chars[i]);
-			g_tree_insert(ambiguous, p, p);
+	for (i = 0; i < G_N_ELEMENTS(_vte_iso2022_unambiguous_ranges); i++) {
+		if ((c >= _vte_iso2022_unambiguous_ranges[i].start) &&
+		    (c <= _vte_iso2022_unambiguous_ranges[i].end)) {
+			return FALSE;
 		}
 	}
-	p = GINT_TO_POINTER(c);
-	return g_tree_lookup(ambiguous, p) == p;
+	if (!ambiguous) {
+		ambiguous = g_hash_table_new (g_direct_hash, g_direct_equal);
+
+		for (i = 0; i < G_N_ELEMENTS(_vte_iso2022_ambiguous_chars); i++) {
+			p = GINT_TO_POINTER(_vte_iso2022_ambiguous_chars[i]);
+			g_hash_table_insert(ambiguous, p, p);
+		}
+	}
+
+	return g_hash_table_lookup(ambiguous, GINT_TO_POINTER(c)) != NULL;
 }
 
+/* If we only have a codepoint, guess what the ambiguous width should be based
+ * on the default region.  Just hope we don't do this too often. */
 static int
-_vte_iso2022_ambiguous_width(void)
+_vte_iso2022_ambiguous_width_guess(void)
 {
 	const char *lang = NULL;
 	int ret = 1;
@@ -355,6 +362,55 @@ _vte_iso2022_ambiguous_width(void)
 	return ret;
 }
 
+/* If we have the encoding, decide how wide an ambiguously-wide character is
+ * based on the encoding.  This is basically what GNU libc does, and it agrees
+ * with my reading of Unicode UAX 11, so.... */
+static int
+_vte_iso2022_ambiguous_width(struct _vte_iso2022_state *state)
+{
+	const char *wide_codelist[] = {
+		"big5",
+		"big5hkscs",
+		"euccn",
+		"eucjp",
+		"euckr",
+		"euctw",
+		"gb18030",
+		"gb2312",
+		"gbk",
+		"tcvn",
+	};
+	int i, j;
+	char codeset[16];
+
+	/* Catch weirdo cases. */
+	if ((state->codeset == NULL) || (state->codeset[0] == '\0')) {
+		return 1;
+	}
+
+	/* Sort-of canonify the encoding name. */
+	i = j = 0;
+	for (i = 0; state->codeset[i] != '\0'; i++) {
+		if (g_ascii_isalnum(state->codeset[i])) {
+			codeset[j++] = g_ascii_tolower(state->codeset[i]);
+		}
+		if (j >= sizeof(codeset) - 1) {
+			break;
+		}
+	}
+	codeset[j] = '\0';
+
+	/* Check for the name in the list. */
+	for (i = 0; i < G_N_ELEMENTS(wide_codelist); i++) {
+		if (strcmp(codeset, wide_codelist[i]) == 0) {
+			return 2;
+		}
+	}
+
+	/* Not in the list => not wide. */
+	return 1;
+}
+
 static GTree *
 _vte_iso2022_map_init(const struct _vte_iso2022_map *map, gssize length)
 {
@@ -810,36 +866,35 @@ _vte_iso2022_state_get_codeset(struct _vte_iso2022_state *state)
 	return state->codeset;
 }
 
-static char *
-_vte_iso2022_better(char *p, char *q)
-{
-	if (p == NULL) {
-		return q;
-	}
-	if (q == NULL) {
-		return p;
-	}
-	return MIN(p, q);
-}
-
 static char *
 _vte_iso2022_find_nextctl(const char *p, size_t length)
 {
 	char *ret;
+	int i;
+
 	if (length == 0) {
 		return NULL;
 	}
-	ret = memchr(p, '\033', length);
-	ret = _vte_iso2022_better(ret, memchr(p, '\n', length));
-	ret = _vte_iso2022_better(ret, memchr(p, '\r', length));
-	ret = _vte_iso2022_better(ret, memchr(p, '\016', length));
-	ret = _vte_iso2022_better(ret, memchr(p, '\017', length));
+
+	for (i = 0; i < length; ++i) {
+		if (p[i] == '\033' ||
+		    p[i] == '\n' ||
+		    p[i] == '\r' ||
+		    p[i] == '\016' ||
+		    p[i] == '\017'
 #ifdef VTE_ISO2022_8_BIT_CONTROLS
-	/* This breaks UTF-8 and other encodings which use the high bits. */
-	ret = _vte_iso2022_better(ret, memchr(p, 0x8e, length));
-	ret = _vte_iso2022_better(ret, memchr(p, 0x8f, length));
+		    /* This breaks UTF-8 and other encodings which
+		     * use the high bits.
+		     */
+  		                 ||
+		    p[i] == 0x8e ||
+		    p[i] == 0x8f
 #endif
-	return ret;
+			) {
+			return (char *)p + i;
+		}
+	}
+	return NULL;
 }
 
 static long
@@ -925,10 +980,14 @@ _vte_iso2022_sequence_length(const unsigned char *nextctl, gsize length)
 				}
 				break;
 #endif
+#if 0
 			case '^':
+				/* ESC ^, the PM.  Search for a string
+				 * terminator. */
+#endif
 			case 'P':
-				/* ESC ^, the PM, or ESC P, the DCS.
-				 * Search for a string terminator. */
+				/* ESC P, the DCS.  Search for a string
+				 * terminator. */
 				for (i = 2; i < length - 1; i++) {
 					if ((nextctl[i] == '\033') &&
 					    (nextctl[i + 1] == '\\')) {
@@ -1155,7 +1214,7 @@ process_8_bit_sequence(struct _vte_iso2022_state *state,
 			width = force_width;
 		} else {
 			if (_vte_iso2022_is_ambiguous(c)) {
-				width = _vte_iso2022_ambiguous_width();
+				width = _vte_iso2022_ambiguous_width(state);
 			}
 		}
 #ifdef VTE_DEBUG
@@ -1178,7 +1237,7 @@ static glong
 process_cdata(struct _vte_iso2022_state *state, guchar *cdata, gsize length,
 	      GArray *gunichars)
 {
-	static int ambiguous_width = 0;
+	int ambiguous_width = 0;
 	glong processed = 0;
 	GTree *map;
 	gint bytes_per_char, force_width, current;
@@ -1192,15 +1251,20 @@ process_cdata(struct _vte_iso2022_state *state, guchar *cdata, gsize length,
 	gpointer p;
 	gboolean single, stop;
 
-	if (ambiguous_width == 0) {
-		ambiguous_width = _vte_iso2022_ambiguous_width();
-	}
+	ambiguous_width = _vte_iso2022_ambiguous_width(state);
 
 	single = (state->override != -1);
 	current = (state->override != -1) ? state->override : state->current;
 	state->override = -1;
 	g_assert((current >= 0) && (current < G_N_ELEMENTS(state->g)));
 
+#ifdef VTE_DEBUG
+	if (_vte_debug_on(VTE_DEBUG_SUBSTITUTION)) {
+		fprintf(stderr, "Current map = %d (%c).\n",
+			current, (state->g[current] & 0xff));
+	}
+#endif
+
 	if (!state->nrc_enabled || (state->g[current] == 'B')) {
 		inbuf = cdata;
 		inbytes = length;
@@ -1210,9 +1274,9 @@ process_cdata(struct _vte_iso2022_state *state, guchar *cdata, gsize length,
 		outbuf = (gunichar*)buf;
 		outbytes = sizeof(gunichar) * length * 2;
 		do {
-			converted = _vte_conv(state->conv,
-					      &inbuf, &inbytes,
-					      (gchar**) &outbuf, &outbytes);
+			converted = _vte_conv_cu(state->conv,
+					         &inbuf, &inbytes,
+					         &outbuf, &outbytes);
 			stop = FALSE;
 			switch (converted) {
 			case ((size_t)-1):
@@ -1265,6 +1329,7 @@ process_cdata(struct _vte_iso2022_state *state, guchar *cdata, gsize length,
 		for (i = 0; &buf[i] < (char*)outbuf; i += sizeof(gunichar)) {
 			c = *(gunichar*)(buf + i);
 			if (c == '\0') {
+				/* Skip the padding character. */
 				continue;
 			}
 			if (_vte_iso2022_is_ambiguous(c)) {
@@ -1370,8 +1435,6 @@ process_control(struct _vte_iso2022_state *state, guchar *ctl, gsize length,
 	if (length >= 1) {
 		switch (ctl[0]) {
 		case '\r':  /* CR */
-			state->current = 0;
-			state->override = -1;
 			c = '\r';
 			g_array_append_val(gunichars, c);
 #ifdef VTE_DEBUG
@@ -1381,8 +1444,6 @@ process_control(struct _vte_iso2022_state *state, guchar *ctl, gsize length,
 #endif
 			break;
 		case '\n':  /* LF */
-			state->current = 0;
-			state->override = -1;
 			c = '\n';
 			g_array_append_val(gunichars, c);
 #ifdef VTE_DEBUG
@@ -1396,7 +1457,7 @@ process_control(struct _vte_iso2022_state *state, guchar *ctl, gsize length,
 			state->override = -1;
 #ifdef VTE_DEBUG
 			if (_vte_debug_on(VTE_DEBUG_SUBSTITUTION)) {
-				fprintf(stderr, "\tSO\n");
+				fprintf(stderr, "\tSO (^N)\n");
 			}
 #endif
 			break;
@@ -1405,7 +1466,7 @@ process_control(struct _vte_iso2022_state *state, guchar *ctl, gsize length,
 			state->override = -1;
 #ifdef VTE_DEBUG
 			if (_vte_debug_on(VTE_DEBUG_SUBSTITUTION)) {
-				fprintf(stderr, "\tSI\n");
+				fprintf(stderr, "\tSI (^O)\n");
 			}
 #endif
 			break;
@@ -1440,6 +1501,21 @@ process_control(struct _vte_iso2022_state *state, guchar *ctl, gsize length,
 						g_array_append_val(gunichars,
 								   c);
 					}
+#ifdef VTE_DEBUG
+					if (_vte_debug_on(VTE_DEBUG_SUBSTITUTION)) {
+						fprintf(stderr, "\t");
+						for (i = 0; i < length; i++) {
+							c = (guchar) ctl[i];
+							fprintf(stderr,
+								"(%s%c)",
+								c < 0x20 ?
+								"^" : "",
+								c < 0x20 ?
+								c : c + 64);
+						}
+						fprintf(stderr, "\n");
+					}
+#endif
 					break;
 				case 'N': /* SS2 */
 					state->override = 2;
@@ -1650,7 +1726,7 @@ _vte_iso2022_process(struct _vte_iso2022_state *state,
 		     GArray *gunichars)
 {
 	GArray *blocks;
-	struct _vte_iso2022_block block;
+	struct _vte_iso2022_block *block = NULL;
 	gboolean preserve_last = FALSE;
 	int i, initial;
 
@@ -1659,19 +1735,19 @@ _vte_iso2022_process(struct _vte_iso2022_state *state,
 	_vte_iso2022_fragment_input(input, blocks);
 
 	for (i = 0; i < blocks->len; i++) {
-		block = g_array_index(blocks, struct _vte_iso2022_block, i);
-		switch (block.type) {
+		block = &g_array_index(blocks, struct _vte_iso2022_block, i);
+		switch (block->type) {
 		case _vte_iso2022_cdata:
 #ifdef VTE_DEBUG
 			if (_vte_debug_on(VTE_DEBUG_SUBSTITUTION)) {
 				int j;
 				fprintf(stderr, "%3ld %3ld CDATA \"%.*s\"",
-					block.start, block.end,
-					(int) (block.end - block.start),
-					input->bytes + block.start);
+					block->start, block->end,
+					(int) (block->end - block->start),
+					input->bytes + block->start);
 				fprintf(stderr, " (");
-				for (j = block.start; j < block.end; j++) {
-					if (j > block.start) {
+				for (j = block->start; j < block->end; j++) {
+					if (j > block->start) {
 						fprintf(stderr, ", ");
 					}
 					fprintf(stderr, "0x%02x",
@@ -1681,14 +1757,14 @@ _vte_iso2022_process(struct _vte_iso2022_state *state,
 			}
 #endif
 			initial = 0;
-			while (initial < block.end - block.start) {
+			while (initial < block->end - block->start) {
 				int j;
 				j = process_cdata(state,
 						  input->bytes +
-						  block.start +
+						  block->start +
 						  initial,
-						  block.end -
-						  block.start -
+						  block->end -
+						  block->start -
 						  initial,
 						  gunichars);
 				if (j == 0) {
@@ -1696,10 +1772,10 @@ _vte_iso2022_process(struct _vte_iso2022_state *state,
 				}
 				initial += j;
 			}
-			if ((initial < block.end - block.start) &&
+			if ((initial < block->end - block->start) &&
 			    (i == blocks->len - 1)) {
 				preserve_last = TRUE;
-				block.start += initial;
+				block->start += initial;
 			} else {
 				preserve_last = FALSE;
 			}
@@ -1708,12 +1784,12 @@ _vte_iso2022_process(struct _vte_iso2022_state *state,
 #ifdef VTE_DEBUG
 			if (_vte_debug_on(VTE_DEBUG_SUBSTITUTION)) {
 				fprintf(stderr, "%3ld %3ld CONTROL ",
-					block.start, block.end);
+					block->start, block->end);
 			}
 #endif
 			process_control(state,
-					input->bytes + block.start,
-					block.end - block.start,
+					input->bytes + block->start,
+					block->end - block->start,
 					gunichars);
 			preserve_last = FALSE;
 			break;
@@ -1721,7 +1797,7 @@ _vte_iso2022_process(struct _vte_iso2022_state *state,
 #ifdef VTE_DEBUG
 			if (_vte_debug_on(VTE_DEBUG_SUBSTITUTION)) {
 				fprintf(stderr, "%3ld %3ld PRESERVE\n",
-					block.start, block.end);
+					block->start, block->end);
 			}
 #endif
 			g_assert(i == blocks->len - 1);
@@ -1733,14 +1809,14 @@ _vte_iso2022_process(struct _vte_iso2022_state *state,
 		}
 	}
 	if (preserve_last && (blocks->len > 0)) {
-		block = g_array_index(blocks, struct _vte_iso2022_block, blocks->len - 1);
+		block = &g_array_index(blocks, struct _vte_iso2022_block, blocks->len - 1);
 #ifdef VTE_DEBUG
 		if (_vte_debug_on(VTE_DEBUG_SUBSTITUTION)) {
-			fprintf(stderr, "Consuming %ld bytes.\n", block.start);
+			fprintf(stderr, "Consuming %ld bytes.\n", block->start);
 		}
 #endif
-		_vte_buffer_consume(input, block.start);
-		g_assert(_vte_buffer_length(input) == block.end - block.start);
+		_vte_buffer_consume(input, block->start);
+		g_assert(_vte_buffer_length(input) == block->end - block->start);
 	} else {
 #ifdef VTE_DEBUG
 		if (_vte_debug_on(VTE_DEBUG_SUBSTITUTION)) {
@@ -1758,7 +1834,7 @@ _vte_iso2022_unichar_width(gunichar c)
 {
 	c = c & ~(VTE_ISO2022_ENCODED_WIDTH_MASK); /* just in case */
 	if (_vte_iso2022_is_ambiguous(c)) {
-		return _vte_iso2022_ambiguous_width();
+		return _vte_iso2022_ambiguous_width_guess();
 	}
 	if (g_unichar_iswide(c)) {
 		return 2;
