@@ -16,12 +16,13 @@
  * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  */
 
-#ident "$Id: vteapp.c,v 1.62 2003/05/22 19:14:49 nalin Exp $"
+#ident "$Id: vteapp.c,v 1.71 2006/01/15 21:09:41 gpastore Exp $"
 
 #include "../config.h"
 
 #include <sys/ioctl.h>
 #include <sys/stat.h>
+#include <errno.h>
 #include <fcntl.h>
 #include <string.h>
 #include <unistd.h>
@@ -49,12 +50,26 @@ window_title_changed(GtkWidget *widget, gpointer win)
 {
 	GtkWindow *window;
 
-	g_return_if_fail(VTE_TERMINAL(widget));
-	g_return_if_fail(GTK_IS_WINDOW(win));
-	g_return_if_fail(VTE_TERMINAL (widget)->window_title != NULL);
+	g_assert(VTE_TERMINAL(widget));
+	g_assert(GTK_IS_WINDOW(win));
+	g_assert(VTE_TERMINAL(widget)->window_title != NULL);
 	window = GTK_WINDOW(win);
 
-	gtk_window_set_title(window, VTE_TERMINAL (widget)->window_title);
+	gtk_window_set_title(window, VTE_TERMINAL(widget)->window_title);
+}
+
+static void
+icon_title_changed(GtkWidget *widget, gpointer win)
+{
+	GtkWindow *window;
+
+	g_assert(VTE_TERMINAL(widget));
+	g_assert(GTK_IS_WINDOW(win));
+	g_assert(VTE_TERMINAL(widget)->icon_title != NULL);
+	window = GTK_WINDOW(win);
+
+	g_message("Icon title changed to \"%s\".\n",
+		  VTE_TERMINAL(widget)->icon_title);
 }
 
 static void
@@ -65,8 +80,8 @@ char_size_changed(GtkWidget *widget, guint width, guint height, gpointer data)
 	GdkGeometry geometry;
 	int xpad, ypad;
 
-	g_return_if_fail(GTK_IS_WINDOW(data));
-	g_return_if_fail(VTE_IS_TERMINAL(widget));
+	g_assert(GTK_IS_WINDOW(data));
+	g_assert(VTE_IS_TERMINAL(widget));
 
 	terminal = VTE_TERMINAL(widget);
 	window = GTK_WINDOW(data);
@@ -230,13 +245,17 @@ static void
 resize_window(GtkWidget *widget, guint width, guint height, gpointer data)
 {
 	VteTerminal *terminal;
-	gint owidth, oheight;
+	gint owidth, oheight, xpad, ypad;
 	if ((GTK_IS_WINDOW(data)) && (width >= 2) && (height >= 2)) {
 		terminal = VTE_TERMINAL(widget);
-		/* Take into account padding and border overhead. */
+		/* Take into account border overhead. */
 		gtk_window_get_size(GTK_WINDOW(data), &owidth, &oheight);
 		owidth -= terminal->char_width * terminal->column_count;
 		oheight -= terminal->char_height * terminal->row_count;
+		/* Take into account padding, which needn't be re-added. */
+		vte_terminal_get_padding(VTE_TERMINAL(widget), &xpad, &ypad);
+		owidth -= xpad;
+		oheight -= ypad;
 		gtk_window_resize(GTK_WINDOW(data),
 				  width + owidth, height + oheight);
 	}
@@ -305,7 +324,7 @@ read_and_feed(GIOChannel *source, GIOCondition condition, gpointer data)
 	char buf[2048];
 	gsize size;
 	GIOStatus status;
-	g_return_val_if_fail(VTE_IS_TERMINAL(data), FALSE);
+	g_assert(VTE_IS_TERMINAL(data));
 	status = g_io_channel_read_chars(source, buf, sizeof(buf),
 					 &size, NULL);
 	if ((status == G_IO_STATUS_NORMAL) && (size > 0)) {
@@ -347,7 +366,12 @@ take_xconsole_ownership(GtkWidget *widget, gpointer data)
 
 	name = g_strdup_printf("MIT_CONSOLE_%s", hostname);
 	atom = gdk_atom_intern(name, FALSE);
+#if GTK_CHECK_VERSION(2,2,0)
+	clipboard = gtk_clipboard_get_for_display(gtk_widget_get_display(widget),
+						  atom);
+#else
 	clipboard = gtk_clipboard_get(atom);
+#endif
 	g_free(name);
 
 	gtk_clipboard_set_with_owner(clipboard,
@@ -358,6 +382,12 @@ take_xconsole_ownership(GtkWidget *widget, gpointer data)
 				     G_OBJECT(widget));
 }
 
+static void
+add_weak_pointer(GObject *object, GtkWidget **target)
+{
+	g_object_add_weak_pointer(object, (gpointer*)target);
+}
+
 int
 main(int argc, char **argv)
 {
@@ -366,7 +396,10 @@ main(int argc, char **argv)
 	const char *background = NULL;
 	gboolean transparent = FALSE, audible = TRUE, blink = TRUE,
 		 debug = FALSE, dingus = FALSE, geometry = TRUE, dbuffer = TRUE,
-		 console = FALSE, scroll = FALSE, keep = FALSE;
+		 console = FALSE, scroll = FALSE, keep = FALSE,
+		 icon_title = FALSE, shell = TRUE, highlight_set = FALSE,
+		 cursor_set = FALSE;
+	VteTerminalAntiAlias antialias = VTE_ANTI_ALIAS_USE_DEFAULT;
 	long lines = 100;
 	const char *message = "Launching interactive shell...\r\n";
 	const char *font = NULL;
@@ -377,10 +410,10 @@ main(int argc, char **argv)
 	int opt;
 	int i, j;
 	GList *args = NULL;
-	GdkColor fore, back, tint;
-	GdkCursor *gumby = NULL, *hand = NULL;
+	GdkColor fore, back, tint, highlight, cursor;
 	const char *usage = "Usage: %s "
 			    "[ [-B image] | [-T] ] "
+			    "[-C] "
 			    "[-D] "
 			    "[-2] "
 			    "[-a] "
@@ -390,11 +423,17 @@ main(int argc, char **argv)
 			    "[-f font] "
 			    "[-g] "
 			    "[-h] "
+			    "[-i] "
 			    "[-k] "
 			    "[-n] "
+			    "[-r] "
+			    "[-s] "
 			    "[-t terminaltype]\n";
 	back.red = back.green = back.blue = 0xffff;
 	fore.red = fore.green = fore.blue = 0x0000;
+	highlight.red = highlight.green = highlight.blue = 0xc000;
+	cursor.red = 0xffff;
+	cursor.green = cursor.blue = 0x8000;
 	tint.red = tint.green = tint.blue = 0;
 	tint = back;
 
@@ -425,9 +464,21 @@ main(int argc, char **argv)
 	g_assert(i < (g_list_length(args) + 2));
 
 	/* Parse some command-line options. */
-	while ((opt = getopt(argc, argv, "B:CDT2abc:df:ghkn:st:w:-")) != -1) {
+	while ((opt = getopt(argc, argv,
+			     "AB:CDST2abc:df:ghkn:rst:w:-")) != -1) {
 		gboolean bail = FALSE;
 		switch (opt) {
+			case 'A':
+				switch (antialias) {
+				case VTE_ANTI_ALIAS_FORCE_DISABLE:
+					antialias = VTE_ANTI_ALIAS_FORCE_ENABLE;
+					break;
+				case VTE_ANTI_ALIAS_USE_DEFAULT:
+				case VTE_ANTI_ALIAS_FORCE_ENABLE:
+					antialias = VTE_ANTI_ALIAS_FORCE_DISABLE;
+					break;
+				}
+				break;
 			case 'B':
 				background = optarg;
 				break;
@@ -437,6 +488,9 @@ main(int argc, char **argv)
 			case 'D':
 				dingus = TRUE;
 				break;
+			case 'S':
+				shell = !shell;
+				break;
 			case 'T':
 				transparent = TRUE;
 				break;
@@ -461,6 +515,12 @@ main(int argc, char **argv)
 			case 'g':
 				geometry = !geometry;
 				break;
+			case 'h':
+				highlight_set = !highlight_set;
+				break;
+			case 'i':
+				icon_title = !icon_title;
+				break;
 			case 'k':
 				keep = !keep;
 				break;
@@ -470,6 +530,9 @@ main(int argc, char **argv)
 					lines = 100;
 				}
 				break;
+			case 'r':
+				cursor_set = !cursor_set;
+				break;
 			case 's':
 				scroll = !scroll;
 				break;
@@ -482,7 +545,6 @@ main(int argc, char **argv)
 			case '-':
 				bail = TRUE;
 				break;
-			case 'h':
 			default:
 				g_print(usage, argv[0]);
 				exit(1);
@@ -499,6 +561,8 @@ main(int argc, char **argv)
 	/* Create a window to hold the scrolling shell, and hook its
 	 * delete event to the quit function.. */
 	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
+	gtk_container_set_resize_mode(GTK_CONTAINER(window),
+				      GTK_RESIZE_IMMEDIATE);
 	g_signal_connect(G_OBJECT(window), "delete_event",
 			 GTK_SIGNAL_FUNC(deleted_and_quit), window);
 
@@ -525,6 +589,10 @@ main(int argc, char **argv)
 	 * window's title. */
 	g_signal_connect(G_OBJECT(widget), "window-title-changed",
 			 G_CALLBACK(window_title_changed), window);
+	if (icon_title) {
+		g_signal_connect(G_OBJECT(widget), "icon-title-changed",
+				 G_CALLBACK(icon_title_changed), window);
+	}
 
 	/* Connect to the "eof" signal to quit when the session ends. */
 	g_signal_connect(G_OBJECT(widget), "eof",
@@ -589,26 +657,33 @@ main(int argc, char **argv)
 	}
 	vte_terminal_set_background_tint_color(VTE_TERMINAL(widget), &tint);
 	vte_terminal_set_colors(VTE_TERMINAL(widget), &fore, &back, NULL, 0);
+	if (highlight_set) {
+		vte_terminal_set_color_highlight(VTE_TERMINAL(widget),
+						 &highlight);
+	}
+	if (cursor_set) {
+		vte_terminal_set_color_cursor(VTE_TERMINAL(widget), &cursor);
+	}
 	if (terminal != NULL) {
 		vte_terminal_set_emulation(VTE_TERMINAL(widget), terminal);
 	}
 
 	/* Set the default font. */
-	if (font != NULL) {
-		vte_terminal_set_font_from_string(VTE_TERMINAL(widget), font);
+	if (font == NULL) {
+		font = "Sans 12";
 	}
+	vte_terminal_set_font_from_string_full(VTE_TERMINAL(widget),
+					       font, antialias);
 
 	/* Match "abcdefg". */
 	vte_terminal_match_add(VTE_TERMINAL(widget), "abcdefg");
 	if (dingus) {
 		i = vte_terminal_match_add(VTE_TERMINAL(widget), DINGUS1);
-		gumby = gdk_cursor_new(GDK_GUMBY);
-		vte_terminal_match_set_cursor(VTE_TERMINAL(widget), i, gumby);
-		gdk_cursor_unref(gumby);
-		hand = gdk_cursor_new(GDK_HAND1);
+		vte_terminal_match_set_cursor_type(VTE_TERMINAL(widget),
+						   i, GDK_GUMBY);
 		i = vte_terminal_match_add(VTE_TERMINAL(widget), DINGUS2);
-		vte_terminal_match_set_cursor(VTE_TERMINAL(widget), i, hand);
-		gdk_cursor_unref(hand);
+		vte_terminal_match_set_cursor_type(VTE_TERMINAL(widget),
+						   i, GDK_HAND1);
 	}
 
 	if (console) {
@@ -655,27 +730,63 @@ main(int argc, char **argv)
 			console = FALSE;
 		}
 	}
+
 	if (!console) {
-		/* Launch a shell. */
-#ifdef VTE_DEBUG
-		if (_vte_debug_on(VTE_DEBUG_MISC)) {
-			vte_terminal_feed(VTE_TERMINAL(widget), message,
-					  strlen(message));
-		}
-#endif
-		vte_terminal_fork_command(VTE_TERMINAL(widget),
-					  command, NULL, env_add,
-					  working_directory,
-					  TRUE, TRUE, TRUE);
-		if (command == NULL) {
-			vte_terminal_feed_child(VTE_TERMINAL(widget),
-						"pwd\n", -1);
+		if (shell) {
+			/* Launch a shell. */
+	#ifdef VTE_DEBUG
+			if (_vte_debug_on(VTE_DEBUG_MISC)) {
+				vte_terminal_feed(VTE_TERMINAL(widget), message,
+						  strlen(message));
+			}
+	#endif
+			vte_terminal_fork_command(VTE_TERMINAL(widget),
+						  command, NULL, env_add,
+						  working_directory,
+						  TRUE, TRUE, TRUE);
+			if (command == NULL) {
+				vte_terminal_feed_child(VTE_TERMINAL(widget),
+							"pwd\n", -1);
+			}
+		} else {
+			long i;
+			i = vte_terminal_forkpty(VTE_TERMINAL(widget),
+						 env_add, working_directory,
+						 TRUE, TRUE, TRUE);
+			switch (i) {
+			case -1:
+				/* abnormal */
+				g_warning("Error in vte_terminal_forkpty(): %s",
+					  strerror(errno));
+				break;
+			case 0:
+				/* child */
+				for (i = 0; ; i++) {
+					switch (i % 3) {
+					case 0:
+					case 1:
+						fprintf(stdout, "%ld\n", i);
+						break;
+					case 2:
+						fprintf(stderr, "%ld\n", i);
+						break;
+					}
+					sleep(1);
+				}
+				_exit(0);
+				break;
+			default:
+				g_print("Child PID is %ld (mine is %ld).\n",
+					(long) i, (long) getpid());
+				/* normal */
+				break;
+			}
 		}
 	}
 
 	/* Go for it! */
-	g_object_add_weak_pointer(G_OBJECT(widget), (gpointer*)&widget);
-	g_object_add_weak_pointer(G_OBJECT(window), (gpointer*)&window);
+	add_weak_pointer(G_OBJECT(widget), &widget);
+	add_weak_pointer(G_OBJECT(window), &window);
 	gtk_widget_show_all(window);
 
 	gtk_main();
